from typing import Any, Callable, Type

import ruamel.yaml

UserType = Any
YamlType = Any

_safe = ruamel.yaml.YAML(typ="safe", pure=True)
_rt = ruamel.yaml.YAML(typ="rt", pure=True)


def register_class(cls: Type) -> None:
    _safe.register_class(cls)
    _rt.register_class(cls)


def add_representer(
    data_type: Type, representer: Callable[[ruamel.yaml.BaseRepresenter, UserType], YamlType]
) -> None:
    _safe.representer.add_representer(data_type, representer)
    _rt.representer.add_representer(data_type, representer)


def add_multi_representer(
    base_data_type: Type,
    multi_representer: Callable[[ruamel.yaml.BaseRepresenter, UserType], YamlType],
) -> None:
    _safe.representer.add_multi_representer(base_data_type, multi_representer)
    _rt.representer.add_multi_representer(base_data_type, multi_representer)


def add_constructor(
    tag: str, constructor: Callable[[ruamel.yaml.BaseConstructor, YamlType], UserType]
) -> None:
    _safe.constructor.add_constructor(tag, constructor)
    _rt.constructor.add_constructor(tag, constructor)


def add_multi_constructor(
    tag_prefix: str,
    multi_constructor: Callable[[ruamel.yaml.BaseConstructor, str, YamlType], UserType],
) -> None:
    _safe.constructor.add_multi_constructor(tag_prefix, multi_constructor)
    _rt.constructor.add_multi_constructor(tag_prefix, multi_constructor)


def _prepare_for_output(d: dict) -> None:
    ruamel.yaml.scalarstring.walk_tree(d)



import contextlib
import dataclasses
import inspect
import logging
import os
import pathlib
import warnings
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import atomicwrites
import pytest

from . import yaml

T = TypeVar("T")


def pytest_addoption(parser):
    parser.addoption(
        "--update-goldens",
        action="store_true",
        default=False,
        help="reset golden master benchmarks",
    )


@pytest.fixture
def golden(request):
    path = None
    try:
        path, func = request.param
    except AttributeError:
        func = request.function

    fixt = GoldenTestFixtureFactory(
        pathlib.Path(request.module.__file__),
        func,
        request.config.getoption("--update-goldens"),
        request.config.getini("enable_assertion_pass_hook"),
    )
    if path is not None:
        fixt = fixt.open(path)
    yield fixt
    fixt.teardown(request.node)


FIXTURE_NAME = golden.__name__
MARKER_NAME = "golden_test"


def pytest_configure(config):
    if not config.getini("enable_assertion_pass_hook"):
        warnings.warn(
            "Add 'enable_assertion_pass_hook=true' to pytest.ini for safer usage of pytest-golden.",
            GoldenTestUsageWarning,
        )
    config.addinivalue_line("markers", MARKER_NAME + "(*file_patterns): TODO")


def _golden_test_marker(*file_patterns: str):
    return file_patterns


class UsageError(Exception):
    pass


class GoldenTestUsageWarning(Warning):
    pass


@dataclasses.dataclass
class GoldenTestFixtureFactory:
    name = FIXTURE_NAME

    path: pathlib.Path
    func: Callable
    update_goldens: bool
    assertions_enabled: bool

    _fixtures = ...  # type: List["GoldenTestFixture"]

    def __post_init__(self):
        self._fixtures = []

    def open(self, path: os.PathLike) -> "GoldenTestFixture":
        kwargs = dataclasses.asdict(self)
        kwargs["path"] = kwargs["path"].parent / path
        fixt = GoldenTestFixture(**kwargs)
        self._fixtures.append(fixt)
        return fixt

    def _add_record(self, r):
        for f in self._fixtures:
            f._add_record(r)

    def teardown(self, item):
        for f in self._fixtures:
            f.teardown(item)


@dataclasses.dataclass
class GoldenTestFixture(GoldenTestFixtureFactory):
    _used_fields = ...  # type: Set[str]
    _records = ...  # type: List[Union["_ComparisonRecord", "_AssertionRecord"]]
    _inputs = ...  # type: Dict[str, Any]

    def __post_init__(self):
        self._used_fields = set()

        # Keep inputs as a separate copy, so if an input gets mutated, it isn't written back.
        with open(self.path, encoding="utf-8") as f:
            self._inputs = yaml._safe.load(f) or {}
        if not isinstance(self._inputs, dict):
            raise UsageError(f"The YAML file '{self.path}' must contain a dict at the top level.")

        if self.update_goldens:
            self.out = GoldenOutputProxy(self)
            self._records = []
        else:
            with open(self.path, encoding="utf-8") as f:
                self.out = yaml._safe.load(f) or {}

    def __getitem__(self, key: str) -> Any:
        self._used_fields.add(key)
        return self._inputs[key]

    def get(self, key: str) -> Optional[Any]:
        self._used_fields.add(key)
        return self._inputs.get(key)

    def _add_record(self, r):
        self._records.append(r)

    def teardown(self, item):
        if not self.update_goldens:
            return

        actual: Dict[str, Union[_AbsentValue, Any]] = {}
        approved_lines: Set[int] = set()
        to_warn: List[Tuple[str, _ComparisonRecord]] = []
        warn = lambda *args: to_warn.append(args)

        for record in reversed(self._records):
            if isinstance(record, _AssertionRecord):
                approved_lines.add(record.lineno)
            elif isinstance(record, _ComparisonRecord):
                comparison = record.comparison
                if record.location.lineno in approved_lines:
                    comparison.approve()
                if self.assertions_enabled and not comparison.approved:
                    warn(
                        f"Comparison to a golden output {record.key!r} outside of an assert is ignored:"
                        f"\n{comparison}",
                        record,
                    )
                    continue
                value = record.other
                if comparison.optional and value is None:
                    value = _AbsentValue()
                if comparison.key in actual and actual[comparison.key] != value:
                    warn(
                        f"Comparison to golden output {comparison.key!r} has gotten conflicting values: "
                        f"{record.other!r} vs {actual[comparison.key]!r}",
                        record,
                    )
                    continue
                actual[record.key] = value

        for msg, record in reversed(to_warn):
            warnings.warn_explicit(
                msg, GoldenTestUsageWarning, record.location.filename, record.location.lineno
            )

        yaml._prepare_for_output(actual)
        with open(self.path, encoding="utf-8") as f:
            outputs = yaml._rt.load(f) or {}
        for k, v in actual.items():
            if isinstance(v, _AbsentValue):
                outputs.pop(k, None)
            else:
                outputs[k] = v

        unused_fields = outputs.keys() - self._used_fields
        if unused_fields:
            f_code = self.func.__code__
            warnings.warn_explicit(
                f"Unused field(s) {', '.join(map(repr, sorted(unused_fields)))} in {item.name}",
                GoldenTestUsageWarning,
                f_code.co_filename,
                f_code.co_firstlineno,
            )
        with atomicwrites.atomic_write(self.path, mode="w", encoding="utf-8", overwrite=True) as f:
            yaml._rt.dump(outputs, f)

    @contextlib.contextmanager
    def may_raise(self, cls: Type[Exception], *, key: str = "exception"):
        try:
            yield
        except cls as e:
            assert self.out.get(key) == {type(e).__name__: str(e)}
        else:
            assert self.out.get(key) == None

    @contextlib.contextmanager
    def capture_logs(
        self,
        loggers: Union[str, Sequence[str]],
        level: int = logging.INFO,
        attributes: Sequence[str] = ("levelname", "getMessage"),
        *,
        key: str = "logs",
    ):
        import testfixtures

        with testfixtures.LogCapture(loggers, attributes=attributes, level=level) as capture:
            yield
        logs = [":".join(log) for log in capture.actual()] or None
        assert self.out.get(key) == logs


@dataclasses.dataclass
class GoldenOutputProxy:
    fixt: GoldenTestFixture

    def __getitem__(self, key: str) -> "GoldenOutput":
        self.fixt._used_fields.add(key)
        return GoldenOutput(self.fixt, key)

    def get(self, key: str) -> "GoldenOutput":
        self.fixt._used_fields.add(key)
        return GoldenOutput(self.fixt, key, optional=True)


@dataclasses.dataclass
class GoldenOutput:
    fixt: GoldenTestFixture
    key: str
    optional: bool = False

    @property
    def value(self):
        return self.fixt[self.key]

    def __eq__(self, other) -> "GoldenComparison":
        if isinstance(other, GoldenOutput):
            raise TypeError("Can't compare two golden output placeholders")
        return GoldenComparison(self.fixt, self.key, other, self.optional)

    def __ne__(self, other) -> "GoldenComparison":
        if isinstance(other, GoldenOutput):
            raise TypeError("Can't compare two golden output placeholders")
        warnings.warn(
            "Only '==' comparison should be used on a golden output",
            GoldenTestUsageWarning,
            stacklevel=2,
        )
        return GoldenComparison(self.fixt, self.key, other, self.optional, eq=False)

    def __str__(self) -> str:
        return f"{self.fixt.name}.out[{self.key!r}]"


@dataclasses.dataclass
class GoldenComparison:
    fixt: GoldenTestFixture
    key: str
    other: Any
    optional: bool
    eq: bool = True
    approved: bool = False

    def __bool__(self) -> bool:
        stack = inspect.stack()
        approved = [
            inspect.unwrap(f).__code__
            for f in (self.fixt.func, GoldenTestFixture.may_raise, GoldenTestFixture.capture_logs)
        ]
        for info in stack:
            if info.frame.f_code in approved:
                self.fixt._add_record(_ComparisonRecord(self, inspect.getframeinfo(info.frame)))
                break
        return self.eq

    def __str__(self) -> str:
        op = "==" if self.eq else "!="
        return f"{self.other!r} {op} {self.fixt.name}.out[{self.key!r}]"

    def approve(self: T) -> T:
        if isinstance(self, GoldenComparison):
            self.approved = True
        return self


@dataclasses.dataclass
class _ComparisonRecord:
    comparison: "GoldenComparison"
    location: inspect.Traceback

    @property
    def key(self) -> str:
        return self.comparison.key

    @property
    def other(self):
        return self.comparison.other


@dataclasses.dataclass
class _AssertionRecord:
    lineno: int


@dataclasses.dataclass
class _AbsentValue:
    def __repr__(self):
        return "<absent>"


def pytest_generate_tests(metafunc):
    item = metafunc.definition
    marker = item.get_closest_marker(MARKER_NAME)
    if not marker:
        return
    patterns = _golden_test_marker(*marker.args, **marker.kwargs)

    f_code = metafunc.function.__code__

    def warn(msg):
        warnings.warn_explicit(
            f"{msg}: {metafunc.function}",
            GoldenTestUsageWarning,
            f_code.co_filename,
            f_code.co_firstlineno,
        )

    if FIXTURE_NAME not in metafunc.fixturenames:
        warn(f"Useless '{MARKER_NAME}' marker on a test without a '{FIXTURE_NAME}' fixture")
        return

    directory = pathlib.Path(metafunc.module.__file__).parent
    paths: Collection[pathlib.Path] = dict.fromkeys(
        path for pattern in patterns for path in directory.glob(pattern)
    )
    if not paths:
        warn(f"The patterns {patterns!r} didn't match anything")
        return

    # `::test_foo[foo/*.yaml]` -> `::test_foo[*.yaml]`
    rel_paths = [path.relative_to(directory) for path in paths]
    skip_parts = None
    if all(
        _removeprefix("test_", path.parts[0]) == _removeprefix("test_", item.originalname)
        for path in rel_paths
    ):
        skip_parts = 1
    ids = ("/".join(path.parts[skip_parts:]) for path in rel_paths)

    metafunc.parametrize(
        FIXTURE_NAME,
        ((path, metafunc.function) for path in paths),
        ids=ids,
        indirect=True,
    )


def _removeprefix(prefix: str, s: str):
    if s.startswith(prefix):
        s = s[len(prefix) :]
    return s


def pytest_assertion_pass(item, lineno, orig, expl):
    fixt = item.funcargs.get(FIXTURE_NAME)
    if isinstance(fixt, GoldenTestFixtureFactory) and fixt.update_goldens:
        fixt._add_record(_AssertionRecord(lineno))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytest_golden']

package_data = \
{'': ['*']}

install_requires = \
['atomicwrites>=1.4.0,<2.0.0',
 'pytest>=6.1.2,<7.0.0',
 'ruamel.yaml>=0.16.12,<0.17.0',
 'testfixtures>=6.15.0,<7.0.0']

extras_require = \
{':python_version < "3.7"': ['dataclasses>=0.7,<0.8']}

entry_points = \
{'pytest11': ['pytest-golden = pytest_golden.plugin']}

setup_kwargs = {
    'name': 'pytest-golden',
    'version': '0.2.1',
    'description': 'Plugin for pytest that offloads expected outputs to data files',
    'long_description': '# pytest-golden\n\nPlugin for [pytest] that offloads expected outputs to data files.\n\n[pytest]: https://pytest.org/\n\n## Usage, in short\n\n(see also: [example/](example/))\n\n[Install the pytest plugin](https://docs.pytest.org/en/latest/plugins.html):\n\n```shell\npip install pytest-golden\n```\n\nCreate a test file (e.g. *tests/test_foo.py*):\n\n```python\n@pytest.mark.golden_test("test_bar/*.yml")\ndef test_bar(golden):\n    assert foo.bar(golden["input"]) == golden.out["output"]\n```\n\nThe wildcard selects the "golden" files which serve as both the input and the expected output for the test. The test is basically parameterized on the files.\n\nCreate one or more of such YAML files (e.g. *tests/test_bar/basic.yml*):\n\n```yaml\ninput: Abc\noutput: Nop\n```\n\nRun `pytest` to execute the test(s).\n\nWhenever the function under test gets changed, its result may change as well, and the test won\'t pass anymore. You can run `pytest --update-goldens` to automatically re-populate the output.\n\n**See [detailed usage](#usage).**\n\n## The case for golden testing\n\nConsider this normal situation when testing a function (e.g. a function to list all words in a sentence).\n\n#### *foo.py*\n\n```python\ndef find_words(text: str) -> list:\n    return text.split()\n```\n\n#### *tests/test_foo.py*\n\n```python\nfrom foo import find_words\n\ndef test_find_words():\n    assert find_words("If at first you don\'t succeed, try, try again.") == [\n        "If", "at", "first", "you", "don\'t", "succeed,", "try,", "try", "again."\n    ]\n```\n\nYou wrote a basic test for that function, but it can be quite tedious to manually write out what the expected output is, especially if the output was something bigger. Sometimes perhaps you\'d resort to just writing a dummy test first and copying the actual output from the failure message. And there\'s nothing really wrong with that, because then you\'d still manually inspect whether the new output is good.\n\n### With golden testing\n\nBut let\'s rewrite this test using "golden testing".\n\n#### *tests/test_foo.py*\n\n```python\nfrom foo import find_words\n\ndef test_find_words(golden):\n    golden = golden.open("test_find_words/test_basic.yml")\n    assert find_words(golden["input"]) == golden.out["output"]\n```\n\nHere `golden["xxx"]` will be a value read directly from the associated file. Let\'s create that (YAML) file:\n\n#### *tests/test_find_words/test_basic.yml*\n\n```yaml\ninput: |-\n  If at first you don\'t succeed, try, try again.\n```\n\nUnlike the input, `golden.out["yyy"]` works a little differently. Normally it will also be just an input for the test, taken from the file (and the assertion will be a completely normal [pytest][] assertion), but in a special "update" mode it will instead accept whatever the result is at runtime and put it back into the "golden" file. Both updating and initially populating the file is done automatically with the command **`pytest --update-goldens`**:\n\n#### *tests/test_find_words/test_basic.yml*\n\n```yaml\ninput: |-\n  If at first you don\'t succeed, try, try again.\noutput:\n- If\n- at\n- first\n- you\n- don\'t\n- succeed,\n- try,\n- try\n- again.\n```\n\nNow, when running just `pytest`, the test will always assert that the result is exactly equal to the expected output. Which is just how unittests work.\n\nNow you can add all of this into your source control system.\n\n### Introducing a change\n\nLet\'s say you\'re not happy that the punctuation gets clumped with the words, so you devise a different implementation for this function.\n\n#### *foo.py*\n\n```python\nimport re\n\ndef find_words(text: str) -> list:\n    return re.findall(r"\\w+", text)\n```\n\nYou also want to add another test case for it:\n\n#### *tests/test_find_words/test_quotation.yml*\n\n```yaml\ninput: |-\n  Dr. King said, \'I have a dream.\'\noutput:\n- Dr\n- King\n- said\n- I\n- have\n- a\n- dream\n```\n\nAnd let\'s just turn this into a *parameterized* golden test (one test generated per each file that matches the wildcard):\n\n#### *tests/test_foo.py*\n\n```python\nimport pytest\nfrom foo import find_words\n\n@pytest.mark.golden_test("test_find_words/*.yml")\ndef test_find_words(golden):\n    assert find_words(golden["input"]) == golden.out["output"]\n```\n\nNow if we run `pytest -v`, we see that all is well with the new test, which gets picked up as `test_find_words[test_quotation.yml]`, but the code changes also made it so the previous test now disagrees! You get a normal failure message from *pytest* itself.\n\nNormally in such situations you\'d go back to the test file and edit the expected output (if you indeed expected it to change). But with this you can instead just run `pytest --update-goldens`, and you\'ll see that instead the "golden" file gets updated by itself (with no test failure). The resulting diff can then still be viewed in your source control system:\n\n```diff\n--- a/tests/test_find_words/test_basic.yml\n+++ b/tests/test_find_words/test_basic.yml\n@@ -5,8 +5,9 @@ output:\n - at\n - first\n - you\n-- don\'t\n-- succeed,\n-- try,\n+- don\n+- t\n+- succeed\n - try\n-- again.\n+- try\n+- again\n```\n\nNow you (and potentially your code reviewers) get to decide whether this diff is an acceptable one, or whether more changes are needed. You can do another iteration on the code, and the unittest will get updated as you go, and you never need to manually edit it -- just visually inspect the changes and check them in.\n\n## Usage\n\n### `golden` fixture\n\nAdd a `golden` parameter to your [pytest][] test function, and it will be passed a `GoldenTestFixtureFactory`.\n\n### class `GoldenTestFixtureFactory`\n\n#### `golden.open(path) -> GoldenTestFixture`\n\nCall this method on the `golden` object to get an actual usable [fixture](#class-goldentestfixture).\n\nThe `path` argument is a path to a file, relative to the calling Python test file. Teardown is done automatically when the test function finishes.\n\n### `@pytest.mark.golden_test(*patterns: str)`\n\nUse this decorator to:\n\n1. avoid having to call `.open` and get a [proper fixture](#class-goldentestfixture) directly as the `golden` argument of your test function and\n2. add parameterization to your "golden" test.\n\nThe `patterns` are one or more [glob patterns](https://docs.python.org/3/library/pathlib.html#pathlib.Path.glob), relative to the calling Python test file. One test will be created for each matched file.\n\n### class `GoldenTestFixture`\n\n#### `golden[input_key: str] -> Any`\n\nGet a value from the associated YAML file, at the top-level key. May raise `KeyError`.\n\n#### `golden.get(input_key: str) -> Optional[Any]`\n\nDitto, but returns `None` if the key is missing.\n\n#### `golden.out[output_key: str] -> Any`\n\n* In normal mode:\n\n  Get a value from the associated YAML file, at the top-level key. May raise `KeyError`.\n\n* If `--update-goldens` flag is passed:\n\n  Get a proxy object for the key, which, upon being compared for equality (and subsequently asserted on), marks that the "golden" file should get an updated value for this top-level key. Such updates get performed upon teardown of the fixture: the original file always gets rewritten once.\n\n#### `golden.out.get(output_key: str) -> Optional[Any]`\n\nDitto, but when compared to `None`, marks the key as deleted from the file, rather than just having the value `None`.\n\n## How to...\n\n### Make a custom type representable in YAML\n\nWe will make these types known to the underlying implementation -- [ruamel.yaml](https://yaml.readthedocs.io/), but let\'s use only the passthrough functions provided by the module `pytest_golden.yaml`. It is best to apply this globally, in *conftest.py*.\n\n```python\nimport pytest_golden.yaml\n\npytest_golden.yaml.register_class(MyClass)\n```\n\n(and see [details for `ruamel.yaml`](https://yaml.readthedocs.io/en/latest/dumpcls.html))\n\nAlternate example if your class is equivalent to a single value:\n\n```python\nclass MyClass:\n    def __init__(self, value: str):\n        self.value = value\n\npytest_golden.yaml.add_representer(MyClass, lambda dumper, data: dumper.represent_scalar("!MyClass", data.value))\npytest_golden.yaml.add_constructor(\'!MyClass\', lambda loader, node: MyClass(node.value))\n```\n\nOr in the particular case of subclassing a standard type, you could just drop the tag altogether and rely on equality to the base type.\n\n```python\nclass MyClass(str):\n    pass\n\npytest_golden.yaml.add_representer(MyClass, lambda dumper, data: dumper.represent_str(data))\n```\n',
    'author': 'Oleh Prypin',
    'author_email': 'oleh@pryp.in',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/oprypin/pytest-golden',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
